;;;; This is the LAML tool that generates the language specific part
;;;; of an XML mirror in Scheme.<p>
;;;; The generated mirrors rely heavily on the common part, which is shared for
;;;; all XML-in-LAML languages. The common part is found in lib/xml-in-laml/xml-in-laml.scm, and
;;;; a <a href="../../../lib/xml-in-laml/man/xml-in-laml.html">XML-in-LAML common library manual</a> is available.<p>
;;;; Let us denote the language by the symbol L. 
;;;; As input this tool takes a parsed DTD, L.dtd, in the Lisp representation
;;;; made by the LAML <a href="../../dtd-parser/man/dtd-parser.html">DTD parser</a>.
;;;; As output the took makes a file L-mirror.scm in the mirror-target-dir.<p>
;;;; The tool is typically driven by a LAML script 
;;;; (such as <a href="../svg10/make-svg10-mirror.laml"> the script for generation of a SVG mirror</a> ) 
;;;; which loads this file. <p>
;;;; The main procedure is generate-mirror. As such, this tool can be identified with the generate-mirror procedure.<p>
;;;; Please consult <a href = "../../../tutorial/xml-in-laml/xml-in-laml.html">"XML mirror in Scheme: XML in LAML"</a> (section 3)
;;;; for a tutorial introduction to the use of the tool.

(lib-load "file-read.scm")

; ---------------------------------------------------------------------------------------------------
;;; Tool parameters - configuration constants.
;;; Please notice that most of these constants must be passed as parameters to the procedure generate-mirror.

;; The name of the generated mirror - a string without initial path and without extension.
;; Must be defined in the LAML script which parameterizes this tool.
(define mirror-name #f)

;; The full path to the parsed DTD file (a lsp file).
;; Must be defined in the LAML script which parameterizes this tool
(define parsed-dtd-path #f)

;; The full path to manually programmed validation predicates.
;; The contents of the file is included in the mirror.
;; If #f, no such predicates will be included.
;; Must be defined in the LAML script which parameterizes this tool
(define manually-programmed-validation-predicates #f)

;; The full path of the directory in which to put the mirror.
;; Must be defined in the LAML script which parameterizes this tool
(define mirror-target-dir #f)

;; Is lib/xml-in-laml/xml-in-laml.scm loaded by the language specific mirror library?
(define auto-lib-loading #f)

;; A boolean variable which controls whether to use the manually programmed content validation
;; predicates, or to use a generic faked checker "faked-generice-content-checker", which always return
;; true. The latter possibility is good for early testing of a mirror.
;; In normal use always #t. If you use #f, be aware that the faked checker is only applied in the cases
;; where we could not automatically synthesize a checking function from the DTD.
(define use-manually-programmed-validation-predicates #t)

;; A variable that controls the naming scheme of the contents model validation predicates.
;; A symbol, either old or normal.
;; The old value is only used for existing stuff. Use normal.
(define validation-predicate-names 'normal)

;; The list of action elements (a list of symbols).
;; The named mirror functions of action elements do not return an AST.
;; The mirror function accessed throught the language map do, however, always return ASTs.
;; Rather, they apply a procedure on the ast.
;; If the element name is N, the action procedure name is N!
(define action-elements '())

; The document type declaration - a string.
; This string is put into the document-type-declaration function, which is called
; by the LAML standard-prolog function.
; (define document-type-declaration-string "")

; ---------------------------------------------------------------------------------------------------
; Scheme name collision.

;; Scheme names - a list of symbols.
;; Think of these as reserved names in Scheme, which we do not want to redefine
;; to mirror functions.
(define reserved-scheme-names 
  '(define do or and else quasiquote begin if quote case lambda set! cond let unquote let* unquote-splicing 
   delay letrec not boolean? eqv? eq? equal? pair? cons set-car! set-cdr! car cdr caar cadr cdar 
   cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr
   cdaaar cdaadr cdadar cdaddr cddaar cddadr cdddar cddddr null? list? list length append rerverse list-tail
   list-ref memq memv member assq assv assoc symbol? symbol->string? string->symbol number? complex? real? rational? 
   integer? exact? inexact? = < > <= >= zero? positive? negative? odd? even? max min + * - / abs quotient remainder
   modulo gcd lcm floor ceiling truncate round number->string string->number char? char=? char<? char>? char<=? char>=?
   char-ci=? char-ci<? char-ci>? char-ci<=? char-ci>=? char-alphabetic? char-numeric? char-whitespace? char-upper-case? 
   char-lower-case? char->integer integer->char char-upcase char-downcase string? make-string string string-length
   string-ref string-set! string=? string-ci=? string<? string>? string<=? string>=? string-ci<? string-ci>? string-ci<=?
   string-ci>=? substring string-append string->list list->string string-copy string-fill! vector? make-vector vector
   vector-length vector-ref vector-set! vector->list list->vector vector-fill! procedure? apply map for-each force
   call-with-current-continuation call-with-input-file call-with-output-file input-port? output-port? current-input-port
   current-output-port with-output-to-file with-input-from-file open-input-file open-output-file close-input-port 
   close-output-port read read-char peek-char eof-object? char-ready? write display newline write-char load transcript-on
   transcript-off rationalize exp log sin cos tan asin acos atan sqrt expt make-rectangular make-polar real-part
   imag-part magnitude angle exact->inexact inexact->exact))

; Is ename (a string or symbol) a reserved scheme name?
(define (is-scheme-name? ename)
 (let ((ename-symbol (as-symbol ename)))
   (turn-into-boolean (memq ename-symbol reserved-scheme-names))))


; ---------------------------------------------------------------------------------------------------
;;; Default language properties.
;;; The default language properties are string-valued variables, which are intended to be set for
;;; each new mirror. The value of these variable are inserted into the generated mirror. This explains
;;; why all values must be strings. This may seem a little strange, but take a look in the beginning of
;;; one of the generated mirror files to see the effect. <p> Notice that lib/xml-in-laml/xml-in-laml.scm contains a number
;;; of configuration options too, together with a number of functions that access/mutate the properties given below.
;;; You should only set those that provide inappropriate defaults.
;;; It means that you typically only have to set of few of the variables in this section when you make a new mirror.


; The default value of xml-check-attributes? The default value is true.
; The value must be a string.
; (define default-xml-check-attributes? "#t")

; The default value of xml-check-contents?  The default value is true.
; The value must be a string.
; (define default-xml-validate-contents? "#t")

;; A boolean value which controls the default value of CDATA transliteration via the HTML/XML charater transformation table.
;; The default value is true.
;; The boolean value must be embedded in a string.
;; This value can be changed after the mirror is generated.
(define default-xml-transliterate-character-data? "#t")

;; The list of elements names (strings) for which the transliteration does note apply.
;; The list value must be embedded in a string.
;; The value of this variable is locked at mirror generation time.
(define default-xml-non-transliteration-elements "'()")

;; The default value of xml-char-transformation-table.
;; The default value is is "html-char-transformation-table".
;; The value must be a string.
(define default-xml-char-transformation-table "html-char-transformation-table")

;; Controls whether to explicitly pass the default dtd attributes in every instance of it.
;; Normally #f, which is the default value. If #t, it will most likely collide with the attribute validation (only
;; one instance of each attribute name is allowed).
;; The boolean value must be embedded a string. 
;; The value of this variable is locked at mirror generation time.
(define default-pass-default-dtd-attributes? "#f")

;; Controls if only string valued attributes are accepted.
;; The default value is true.
;; The boolean value must be embedded in a string.
(define default-xml-accept-only-string-valued-attributes? "#t")

;; The default value of document-type-declaration.
;; The value must be a string. The default default is the empty string with the meaning
;; 'no attribute type declaration'.
;; The value of this variable is locked at mirror generation time.
(define default-document-type-declaration "\"\"")


;; Controls the presence of white space markers in the ast.
;; The default value is true.
;; The boolean value must be a string.
(define default-xml-represent-white-space "#t")


; ---------------------------------------------------------------------------------------------------
; Tool encapsulation functions:

; Return the name of the validation procudure of the element name in language.
(define (name-of-validation-procedure language-name name)
 (string-append (as-string name) "-" (as-string language-name) "-" "laml-validate" "!"))

; Return the name of the content model validation predicate.
; These are the predicates some of whic we need to write manually.
(define (name-of-content-validation-predicate language-name name)
 (cond ((eq? validation-predicate-names 'old) 
         (string-append (as-string name) "-" "checker"))
       ((eq? validation-predicate-names 'normal)
         (string-append (as-string name) "-" (as-string language-name) "-" "checker" "?"))
       (else (laml-error "name-of-content-validation-predicate: unknown value of validation-predicate-names: " validation-predicate-names))))

; Return the name of an action procedure given the element-name.
(define (action-procedure-name element-name)  
  (string-append (as-string element-name) "!"))


; ---------------------------------------------------------------------------------------------------

;;; The main tool procedure.
;;; The main tool procedure is generate-mirror. After setting the parameters call this procedure
;;; which will make the XML mirror.


;; The main tool procedure which makes the mirror of XML in LAML.
;; The procedure writes a text file in mirror-destination-path.
;; .parameter parsed-dtd-path the full path to the parsed dtd (a lsp file).
;; .parameter mirror-destination-path full path to the file on which to write the mirror.
;; .parameter validation-predicate-path the full path the manually programmed validation precidates, which we include as source text in the mirror.
;; .parameter language-name the name of language which we mirror in LAML (a symbol).
(define (generate-mirror parsed-dtd-path mirror-destination-path validation-predicate-path language-name . optional-parameter-list)
   (set! manually-programmed-validation-predicates validation-predicate-path)
   (let* ((dtd-list (file-read parsed-dtd-path))
          (elements (filter (lambda (el) (eq? (car el) 'element)) dtd-list))
          (attributes (filter (lambda (el) (eq? (car el) 'attribute)) dtd-list))
          (sorted-attributes   ; attributes sorted in the same order as elements
            (sort-list
              attributes
              (generate-leq (map (compose as-symbol cadr) elements) (compose as-symbol cadr))))
         )
  
  
  
    (write-text-file
      (string-append
       "; This file is generated by an LAML script based on the LAML tool tools/xml-in-laml/xml-in-laml.scm. DO NOT EDIT!" CR CR
  
       (if auto-lib-loading
           (string-append
  	  ";;; Loading the XML-in-LAML stuff common for all languages:" CR
  	  "(load (string-append laml-dir \"lib/xml-in-laml/xml-in-laml.scm\"))" CR CR)
           (string-append
            "; lib/xml-in-laml/xml-in-laml.scm is not loaded here. " CR
            "; You must load it your self prior to the loading of this file." CR CR))

       (language-property-definitions language-name)  CR CR
  
       "; Empty temporary language map" CR
       (as-string `(set! temp-language-map '())) CR CR
  
       CR CR 
       ";;; The validation procedures" CR CR
       (accumulate-right
         string-append
         ""
        (map2 
          (lambda (element attribute)
            (make-xml-validation-procedure element attribute language-name))
          elements sorted-attributes))
  
       ";;; Make and put XML mirror functions in the temporary language map:" CR
       (accumulate-right
         string-append
         ""
        (map2 
          (lambda (element attribute)
            (make-xml-mirror-function element attribute language-name))
          elements sorted-attributes))
  
       "; Register the name of the language:" CR
       (as-string `(register-xml-in-laml-language (quote ,language-name) temp-language-map))   CR CR
  
       "; Define the language variable" CR
       (as-string `(define ,language-name (activator-via-language-map (quote ,language-name))))   CR CR
  
  
       CR CR
       
       (if (and manually-programmed-validation-predicates (file-exists? manually-programmed-validation-predicates))
           (string-append
             (string-append ";;; Manually programmed validation predicates" CR CR)
             (read-text-file manually-programmed-validation-predicates))
           "")
  
  
  
      ) 
      mirror-destination-path)
  
     'DONE))


(define (language-property-definitions language-name)
 (string-append
   (definition-for language-name "xml-transliterate-character-data?" default-xml-transliterate-character-data?)
   (definition-for language-name "xml-char-transformation-table" default-xml-char-transformation-table)
   (definition-for language-name "xml-non-transliteration-elements" default-xml-non-transliteration-elements)

   (definition-for language-name "xml-pass-default-dtd-attributes?" default-pass-default-dtd-attributes?)
   (definition-for language-name "xml-accept-only-string-valued-attributes?" default-xml-accept-only-string-valued-attributes?)
   (definition-for language-name "xml-document-type-declaration" default-document-type-declaration)
   (definition-for language-name "xml-represent-white-space?" default-xml-represent-white-space)))



; Generate a textual definition, which can be inserted into the generated mirror.
(define (definition-for language-name variable-name value-string)
 (let ((SP " "))
  (string-append
    "(define" SP 
              (as-string language-name) "-" (as-string variable-name) SP
              value-string ")" CR)))
              


;; Construct and return a mirror function.
;; .parameter element    The parsed element structure, as produced by the LAML DTD parser.
;; .parameter attribute  The parsed attribute structure, as produced by the LAML DTD parser.
;; .parameter language-name The name of the XML language, to which this mirror function belongs.
(define (make-xml-mirror-function element attribute language-name)
 (let* ((ename (el-name element))
        (content-model (el-content-model element))
        (scheme-collision? (is-scheme-name? ename))
        (comment (el-comment element))
        (aname (at-name attribute))
        (alist (at-list attribute))
        (validation-procedure-name (name-of-validation-procedure language-name ename))
        (kind (cond ((single-tag? content-model) 'single)
                    ((double-tag? content-model) 'double)
                    (else (laml-error "make-xml-mirror-function: Neither single or double: " ename))))
        (action-element? (turn-into-boolean (memq (as-symbol ename) action-elements)))
        (default-dtd-attributes (extract-defaulted-attributes attribute))
       )
    (if scheme-collision?
        (display-warning 
         (string-append ename ": Element name collides with reserved Scheme name. No named mirror function defined.")))

    (if (eq? (as-symbol language-name) (as-symbol ename))
        (display-warning 
          (string-append ename ": Element name collides with the name of the XML language. Chose another language name!" CR
                               "No named mirror function available.")))

    (string-append 
     (make-xml-mirror-function-help kind ename default-dtd-attributes validation-procedure-name scheme-collision?
                                    language-name action-element?)
     CR CR)
  )
)
   

(define (make-xml-mirror-function-help kind ename default-dtd-attributes validation-procedure scheme-collision? language-name action-element?)
 (let ((action-procedure (if action-element? (action-procedure-name ename) #f)))
  (string-append
   (as-string-1
    `(set! temp-mirror-function
      (generate-xml-mirror-function ,validation-procedure ,(string-it ename) (quote ,default-dtd-attributes)
                                    (quote ,kind) (quote ,language-name) #f #f))) CR

   (as-string
    `(set! temp-language-map (put-mirror-function temp-language-map ,(string-it ename) temp-mirror-function))) CR

   (if (not scheme-collision?)
       (string-append
	(as-string-1
	 `(set! temp-mirror-function
	   (generate-xml-mirror-function ,validation-procedure ,(string-it ename) (quote ,default-dtd-attributes)
                                         (quote ,kind) (quote ,language-name) #t ,action-procedure))) 
           CR)
       "")

   (if (not scheme-collision?)
       (as-string
	`(define ,ename temp-mirror-function)
        )
       (string-append "; No mirror of " (as-string ename) " defined due to name clash with reversed Scheme name." ))

   )
  )
)

;; Make and return a validation procedure.
;; .parameter element    The parsed element structure, as produced by the LAML DTD parser.
;; .parameter attribute  The parsed attribute structure, as produced by the LAML DTD parser.
;; .parameter language-name The name of the XML language, to which this mirror function belongs.
(define (make-xml-validation-procedure element attribute language-name)
 (let* ((ename (el-name element))
        (content-model (el-content-model element))
        (comment (el-comment element))
        (aname (at-name attribute))
        (alist (at-list attribute))
        (val-proc-name (name-of-validation-procedure language-name ename))
       )
 (if (not (equal? ename aname))
     (error 
      (string-append 
        "make-html-basic-mirror-function: element name and attribute name must be the same. " (as-string ename) " " (as-string aname)  CR
        "For each element definition there must be an explicit attribute definition.")))
 
 (string-append 
     (cond 
           ((single-tag? content-model)  ; important first to test for single-tag
            (make-single-tag-xml-validation-procedure val-proc-name ename comment alist content-model language-name))
           ((double-tag? content-model)   
            (make-double-tag-xml-validation-procedure val-proc-name ename comment alist content-model language-name))
           (else (error (string-append "make-xml-validation-procedure: " ename " not single-tag or double-tag."))))
     CR CR)
 )
)

(define (make-single-tag-xml-validation-procedure val-proc-name tag-name comment attribute-list rhs language)
 (let* (
        (lc-tag-name tag-name)
        (lc-tag-name-string (string-it lc-tag-name))
        (attribute-list-1 (reorder-attributes attribute-list))
        (number-of-required-attributes (length (filter required-attribute? attribute-list-1))) ; could be made more efficient...
        (stringed-attribute-list (map (lambda (triple) (map string-it-1 triple)) attribute-list-1))
       )
  (string-append
   (as-string
    `(define (,val-proc-name el-name attributes contents overlap-check?)  ; earlier (,val-proc-name . attributes)
       (let ((attributes-of-elements (quote ,stringed-attribute-list))
             (req-n ,number-of-required-attributes)
            )
         (xml-check-for-empty-contents! contents ,(string-it lc-tag-name))

         (if (and overlap-check? xml-check-language-overlap?)
             (check-language-overlap! (as-symbol ,(string-it lc-tag-name))))

         (if xml-check-attributes? (xml-check-attributes! attributes attributes-of-elements req-n ,(string-it lc-tag-name)))))
   )
  )
 )
)


(define (make-double-tag-xml-validation-procedure val-proc-name tag-name comment attribute-list rhs language)
 (let* (
        (lc-tag-name tag-name)
        (lc-tag-name-string (string-it lc-tag-name))
        (attribute-list-1 (reorder-attributes attribute-list))
        (number-of-required-attributes (length (filter required-attribute? attribute-list-1))) ; could be made more efficient...
        (par (lambda(str) ""))
        (com-format (lambda(str) (par (if (empty-string? str) ""  (strip-initial-spaces str)))))
        (stringed-attribute-list (map (lambda (triple) (map string-it-1 triple)) attribute-list-1))
        (stringed-constituent-list 
          (if (and (list? rhs) (not (null? rhs)))
              (cons (car rhs) (map string-or-quote-it (cdr rhs)))
              #f))
        (validator
          (cond (stringed-constituent-list stringed-constituent-list )
                ((symbol? rhs) rhs)
                (else (if use-manually-programmed-validation-predicates
			  (name-of-content-validation-predicate language lc-tag-name)
			  "faked-generice-content-checker"))))
       )
  (string-append
   (as-string-1
    `(define (,val-proc-name el-name attributes contents overlap-check?)
       (let ((attributes-of-elements (quote ,stringed-attribute-list))
             (req-n ,number-of-required-attributes)
            )
         (if (and overlap-check? xml-check-language-overlap?)
             (check-language-overlap! (as-symbol ,(string-it lc-tag-name))))

         (if xml-check-attributes?
             (xml-check-attributes! attributes attributes-of-elements req-n ,(string-it lc-tag-name)))

         (if xml-validate-contents? 
             (validate-contents! 
                contents
                ,validator
              ,(string-it lc-tag-name)))

     ))
    )
   )
 )
)

(define (string-or-quote-it x)
  (cond ((string? x) (string-it x))
        ((symbol? x) (list 'quote x))))



; Re-arrange attribute-list such that the required attributes comes first.
(define (reorder-attributes attribute-list)
  (let* (
         (required-attributes (filter required-attribute? attribute-list))
         (optional-attributes (filter optional-attribute? attribute-list)))
    (append required-attributes optional-attributes)))



; Element and attribute selectors - in relation to Lisp representation of DTD information.

(define el-name (make-selector-function 2))
(define el-start-tag (make-selector-function 3))
(define el-end-tag (make-selector-function 4))
(define el-content-model (make-selector-function 5))
(define el-comment (make-selector-function 6))

; Selectors on a parsed DTD attribute structure such as the list
; (attribute "el-name" (("id" "CDATA" "#IMPLIED") ("margin" "CDATA" "#IMPLIED")))
(define at-name (make-selector-function 2))
(define at-list (make-selector-function 3))

; Selectors on a single DTD attribute, such as 
; ("id" "CDATA" "#IMPLIED") or
; ("id" "CDATA" "xxx" "#FIXED") or
(define att-name (make-selector-function 1))
(define att-type (make-selector-function 2))
(define att-status-and-default (make-selector-function 3))

(define att-fixed? 
  (lambda (a) 
    (if (= (length a) 4)
        (equal? (fourth a) "#FIXED")
        #f)))


; Single double tag distinctions.

; Does the content-model of an element indicate an empty element (single-tag).
(define (single-tag? content-model)
  (equal? content-model "EMPTY"))

; Does the content-model of an element indicate a non-empty element (double-tag).
(define (double-tag? content-model)
  (not (equal? content-model "EMPTY")))

(define (required-attribute? attr-structure)
  (equal? (att-status-and-default attr-structure) "#REQUIRED"))

(define (optional-attribute? attr-structure)
  (not (equal? (att-status-and-default attr-structure) "#REQUIRED")))

(define (defaulted-attribute? attr-structure)
 (let ((att-status-default (att-status-and-default attr-structure)))
  (not 
    (or (equal? att-status-default "#REQUIRED")
        (equal? att-status-default "#IMPLIED")))))

; Return an attribute list - a property list - of attribute name / attribute default value.
; .parameter dtd-attribute-structure the attribute structure of a parsed DTD structure.
(define (extract-defaulted-attributes dtd-attribute-structure)
 (let ((dtd-attr-list (at-list dtd-attribute-structure))
       (dtd-att-to-runtime-att 
            (lambda (dtd-att)
              (cons (att-name dtd-att) (string-it (att-status-and-default dtd-att)))))
      )
   (alist-to-propertylist (map dtd-att-to-runtime-att (filter defaulted-attribute? dtd-attr-list)))))
  



; A more general version of string-it, which is able to recursively apply string-it on lists of strings.
(define (string-it-1 x)
  (cond ((string? x) (string-it x))
        ((list? x) (map string-it-1 x))
        (else (error "string-it-1: Applied on unsupported element"))))

; A variant of as-string, which translates booleans to #t and #f instead of true and false
(define (as-string-1 x)
  (cond ((number? x) (number->string x))
        ((symbol? x) (symbol->string x))
        ((string? x) x)
        ((boolean? x) 
            (if x "#t" "#f"))   ; true and false in the normal version 
        ((char? x) (char->string x))
        ((list? x)
            (string-append "(" 
               (string-merge (map as-string-1 x) (make-list (- (length x) 1) " "))
               ")"))
        ((pair? x)
            (string-append "(" 
               (apply string-append
                  (map (lambda (y) (string-append (as-string-1 y) " ")) (proper-part x))
               )
               " . " (as-string-1 (first-improper-part x))
               ")"))
        (else "??")))

; Move to general...
(define CR (as-string #\newline))
